@@ -6,7 +6,8 @@
 ;;; specific commands such as show type signature, show info, haskell process
 ;;; commands and etc.
 
-;; Copyright (c) 2014 Chris Done. All rights reserved.
+;; Copyright © 2014 Chris Done.  All rights reserved.
+;;             2016 Arthur Fayzrakhmanov
 
 ;; This file is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -25,6 +26,7 @@
 
 (require 'cl-lib)
 (require 'etags)
+(require 'haskell-mode)
 (require 'haskell-compat)
 (require 'haskell-process)
 (require 'haskell-font-lock)
@@ -33,6 +35,7 @@
 (require 'haskell-presentation-mode)
 (require 'haskell-utils)
 (require 'highlight-uses-mode)
+(require 'haskell-cabal)
 
 ;;;###autoload
 (defun haskell-process-restart ()
@@ -164,8 +167,7 @@ MODULE-BUFFER is the actual Emacs buffer of the module being loaded."
   "History list for session targets.")
 
 (defun haskell-process-hayoo-ident (ident)
-  ;; FIXME Obsolete doc string, CALLBACK is not used.
-  "Hayoo for IDENT, return a list of modules asyncronously through CALLBACK."
+  "Hayoo for IDENT, return a list of modules"
   ;; We need a real/simulated closure, because otherwise these
   ;; variables will be unbound when the url-retrieve callback is
   ;; called.
@@ -276,7 +278,7 @@ Prompts for an arbitrary regexp given a prefix arg PROMPT."
                  (read-from-minibuffer "Look for: ")
                (haskell-ident-at-point))))
     (rgrep sym
-           "*.hs" ;; TODO: common Haskell extensions.
+           "*.hs *.lhs *.hsc *.chs *.hs-boot *.lhs-boot"
            (haskell-session-current-dir (haskell-interactive-session)))))
 
 ;;;###autoload
@@ -309,7 +311,6 @@ If PROMPT-VALUE is non-nil, request identifier via mini-buffer."
 
 ;;;###autoload
 (defun haskell-process-do-type (&optional insert-value)
-  ;; FIXME insert value functionallity seems to be missing.
   "Print the type of the given expression.
 
 Given INSERT-VALUE prefix indicates that result type signature
@@ -355,7 +356,9 @@ position with `xref-pop-marker-stack'."
         (haskell-mode-handle-generic-loc loc)
       (call-interactively 'haskell-mode-tag-find))
     (unless (equal initial-loc (point-marker))
-      (xref-push-marker-stack initial-loc))))
+      (save-excursion
+        (goto-char initial-loc)
+        (xref-push-marker-stack)))))
 
 ;;;###autoload
 (defun haskell-mode-goto-loc ()
@@ -456,7 +459,12 @@ Returns:
   "Either jump to or echo a generic location LOC.
 Either a file or a library."
   (cl-case (car loc)
-    (file (haskell-mode-jump-to-loc (cdr loc)))
+    (file (progn
+              (find-file (elt loc 1))
+              (goto-char (point-min))
+              (forward-line (1- (elt loc 2)))
+              (goto-char (+ (line-beginning-position)
+                            (1- (elt loc 3))))))
     (library (message "Defined in `%s' (%s)."
                       (elt loc 2)
                       (elt loc 1)))
@@ -520,8 +528,8 @@ of which the latter defaults to the current buffer."
   (or (haskell-session-get session 'current-dir)
       (haskell-session-get session 'cabal-dir)
       (if (buffer-file-name buffer)
-	  (file-name-directory (buffer-file-name buffer))
-	  "~/")))
+          (file-name-directory (buffer-file-name buffer))
+          "~/")))
 
 (defun haskell-session-prompt-set-current-dir (session &optional use-default)
   "Prompt for the current directory.
@@ -530,8 +538,8 @@ Return current working directory for SESSION."
     (haskell-session-set-current-dir
      session
      (if use-default
-	 default
-	 (haskell-utils-read-directory-name "Set current directory: " default))))
+         default
+         (haskell-utils-read-directory-name "Set current directory: " default))))
   (haskell-session-get session 'current-dir))
 
 (defun haskell-process-change-dir (session process dir)
@@ -636,7 +644,7 @@ happened since function invocation)."
                (min-pos (caar pos-reg))
                (max-pos (cdar pos-reg))
                (sig (haskell-utils-reduce-string response))
-               (res-type (haskell-utils-parse-repl-response sig)))
+               (res-type (haskell-utils-repl-response-error-status sig)))
 
           (cl-case res-type
             ;; neither popup presentation buffer
@@ -682,7 +690,9 @@ happened since function invocation)."
 
             (haskell-utils-async-stop-watching-changes init-buffer))))))))
 
-;;;###autoload
+(make-obsolete 'haskell-process-generate-tags
+               'haskell-mode-generate-tags
+               "2016-03-14")
 (defun haskell-process-generate-tags (&optional and-then-find-this-tag)
   "Regenerate the TAGS table.
 If optional AND-THEN-FIND-THIS-TAG argument is present it is used with
@@ -693,20 +703,13 @@ function `xref-find-definitions' after new table was generated."
      process
      (make-haskell-command
       :state (cons process and-then-find-this-tag)
-      :go (lambda (state)
-            (if (eq system-type 'windows-nt)
-                (haskell-process-send-string
-                 (car state)
-                 (format ":!hasktags --output=\"%s\\TAGS\" -x -e \"%s\""
-                            (haskell-session-cabal-dir (haskell-process-session (car state)))
-                            (haskell-session-cabal-dir (haskell-process-session (car state)))))
-              (haskell-process-send-string
-               (car state)
-               (format ":!cd %s && %s | %s"
-                       (haskell-session-cabal-dir
-                        (haskell-process-session (car state)))
-                       "find . -type f \\( -name '*.hs' -or -name '*.lhs' -or -name '*.hsc' \\) -not \\( -name '#*' -or -name '.*' \\) -print0"
-                       "xargs -0 hasktags -e -x"))))
+      :go
+      (lambda (state)
+        (let* ((process (car state))
+               (cabal-dir (haskell-session-cabal-dir
+                           (haskell-process-session process)))
+               (command (haskell-cabal--compose-hasktags-command cabal-dir)))
+          (haskell-process-send-string process command)))
       :complete (lambda (state _response)
                   (when (cdr state)
                     (let ((tags-file-name
@@ -737,39 +740,53 @@ inferior GHCi process."
   (interactive)
   (let ((session (haskell-interactive-session))
         (changed nil))
-    (if (null (haskell-session-get session
-                                   'ignored-files))
+    (if (null (haskell-session-get session 'ignored-files))
         (message "Nothing to unignore!")
-      (cl-loop for file in (haskell-session-get session
-                                                'ignored-files)
-               do (cl-case (read-event
-                            (propertize (format "Set permissions? %s (y, n, v: stop and view file)"
-                                                file)
-                                        'face 'minibuffer-prompt))
-                    (?y
-                     (haskell-process-unignore-file session file)
-                     (setq changed t))
-                    (?v
-                     (find-file file)
-                     (cl-return))))
-      (when (and changed
-                 (y-or-n-p "Restart GHCi process now? "))
-        (haskell-process-restart)))))
+      (cl-loop for file in (haskell-session-get session 'ignored-files)
+               do
+               (haskell-mode-toggle-interactive-prompt-state)
+               (unwind-protect
+                   (progn
+                     (cl-case
+                         (read-event
+                          (propertize
+                           (format "Set permissions? %s (y, n, v: stop and view file)"
+                                   file)
+                           'face
+                           'minibuffer-prompt))
+                       (?y
+                        (haskell-process-unignore-file session file)
+                        (setq changed t))
+                       (?v
+                        (find-file file)
+                        (cl-return)))
+                     (when (and changed
+                                (y-or-n-p "Restart GHCi process now? "))
+                       (haskell-process-restart)))
+                 ;; unwind
+                 (haskell-mode-toggle-interactive-prompt-state t))))))
 
 ;;;###autoload
 (defun haskell-session-change-target (target)
   "Set the build TARGET for cabal REPL."
   (interactive
    (list
-     (completing-read "New build target: " (haskell-cabal-enum-targets)
-		      nil nil nil 'haskell-cabal-targets-history)))
+    (completing-read "New build target: "
+                     (haskell-cabal-enum-targets)
+                     nil
+                     nil
+                     nil
+                     'haskell-cabal-targets-history)))
   (let* ((session haskell-session)
          (old-target (haskell-session-get session 'target)))
     (when session
       (haskell-session-set-target session target)
-      (when (and (not (string= old-target target))
-                 (y-or-n-p "Target changed, restart haskell process?"))
-        (haskell-process-start session)))))
+      (when (not (string= old-target target))
+        (haskell-mode-toggle-interactive-prompt-state)
+        (unwind-protect
+            (when (y-or-n-p "Target changed, restart haskell process?")
+              (haskell-process-start session)))
+        (haskell-mode-toggle-interactive-prompt-state t)))))
 
 ;;;###autoload
 (defun haskell-mode-stylish-buffer ()
@@ -792,8 +809,8 @@ output.  If CMD fails the buffer remains unchanged."
                     (setq str (replace-match "" t t str)))
                   str))
          (_errout (lambda (fmt &rest args)
-		    (let* ((warning-fill-prefix "    "))
-		      (display-warning cmd (apply 'format fmt args) :warning))))
+                    (let* ((warning-fill-prefix "    "))
+                      (display-warning cmd (apply 'format fmt args) :warning))))
          (filename (buffer-file-name (current-buffer)))
          (cmd-prefix (replace-regexp-in-string " .*" "" cmd))
          (tmp-file (make-temp-file cmd-prefix))
@@ -803,8 +820,8 @@ output.  If CMD fails the buffer remains unchanged."
                                 (haskell-session-cabal-dir haskell-session)
                               default-directory))
          (_errcode (with-temp-file tmp-file
-		     (call-process cmd filename
-				   (list (current-buffer) err-file) nil)))
+                     (call-process cmd filename
+                                   (list (current-buffer) err-file) nil)))
          (stderr-output
           (with-temp-buffer
             (insert-file-contents err-file)
