@@ -49,7 +49,6 @@
   :group 'haskell
   :prefix "haskell-indentation-")
 
-;;;###autoload
 (defcustom haskell-indentation-indent-leftmost t
   "Indent to the left margin after certain keywords.
 For example after \"let .. in\", \"case .. of\").  If set to t it
@@ -59,47 +58,42 @@ positions are allowed."
   :type 'symbol
   :group 'haskell-indentation)
 
-;;;###autoload
 (defcustom haskell-indentation-layout-offset 2
   "Extra indentation to add before expressions in a Haskell layout list."
   :type 'integer
   :group 'haskell-indentation)
 
-;;;###autoload
 (defcustom haskell-indentation-starter-offset 2
   "Extra indentation after an opening keyword (e.g. \"let\")."
   :type 'integer
   :group 'haskell-indentation)
 
-;;;###autoload
 (defcustom haskell-indentation-left-offset 2
   "Extra indentation after an indentation to the left (e.g. after \"do\")."
   :type 'integer
   :group 'haskell-indentation)
 
-;;;###autoload
 (defcustom  haskell-indentation-ifte-offset 2
   "Extra indentation after the keywords \"if\", \"then\", or \"else\"."
   :type 'integer
   :group 'haskell-indentation)
 
-;;;###autoload
 (defcustom haskell-indentation-where-pre-offset 2
   "Extra indentation before the keyword \"where\"."
   :type 'integer
   :group 'haskell-indentation)
 
-;;;###autoload
 (defcustom haskell-indentation-where-post-offset 2
   "Extra indentation after the keyword \"where\"."
   :type 'integer
   :group 'haskell-indentation)
 
 (defconst haskell-indentation-mode-map
-  (let ((keymap (make-sparse-keymap)))
-    (define-key keymap (kbd "RET") 'haskell-indentation-newline-and-indent)
-    (define-key keymap (kbd "<backtab>") 'haskell-indentation-indent-backwards)
-    keymap))
+  (let ((map (make-sparse-keymap)))
+    (define-key map (kbd "RET") 'haskell-indentation-newline-and-indent)
+    (define-key map (kbd "<backtab>") 'haskell-indentation-indent-backwards)
+    map)
+  "Keymap for `haskell-indentation-mode'.")
 
 ;;;###autoload
 (define-minor-mode haskell-indentation-mode
@@ -135,8 +129,7 @@ set and deleted as if they were real tabs."
 (defvar haskell-literate) ; defined in haskell-mode.el
 
 (defun haskell-indentation-bird-p ()
-  "Return t if this is a literate Haskell buffer in bird style,
-NIL otherwise."
+  "Return t if this is a literate Haskell buffer in bird style, NIL otherwise."
   (eq haskell-literate 'bird))
 
 ;;----------------------------------------------------------------------------
@@ -217,7 +210,7 @@ negative ARG.  Handles bird style literate Haskell too."
 (defun haskell-indentation-next-indentation (col indentations &optional nofail)
   "Find the leftmost indentation which is greater than COL.
 Indentations are taken from INDENTATIONS, which should be a
-list. Return the last indentation if there are no bigger ones and
+list.  Return the last indentation if there are no bigger ones and
 NOFAIL is non-NIL."
   (when (null indentations)
     (error "haskell-indentation-next-indentation called with empty list"))
@@ -226,7 +219,9 @@ NOFAIL is non-NIL."
         (car (last indentations)))))
 
 (defun haskell-indentation-previous-indentation (col indentations &optional nofail)
-  "Find the rightmost indentation which is less than COL."
+  "Find the rightmost indentation less than COL from INDENTATIONS.
+When no indentations are less than COL, return the rightmost indentation
+if NOFAIL is non-nil, or nil otherwise."
   (when (null indentations)
     (error "haskell-indentation-previous-indentation called with empty list"))
   (let ((rev (reverse indentations)))
@@ -234,8 +229,6 @@ NOFAIL is non-NIL."
         (when nofail
           (car rev)))))
 
-(defvar haskell-indentation-dyn-first-position nil
-  "") ; FIXME
 (defvar haskell-indentation-dyn-last-direction nil
   "") ; FIXME
 (defvar haskell-indentation-dyn-last-indentations nil
@@ -251,31 +244,25 @@ indentation points to the right, we switch going to the left."
   ;; try to repeat
   (when (not (haskell-indentation-indent-line-repeat))
     (setq haskell-indentation-dyn-last-direction nil)
-    ;; do nothing if we're inside a string or comment
-    (unless (save-excursion
-              (beginning-of-line)
-              (nth 8 (syntax-ppss)))
-      ;; parse error is intentionally not cought here, it may come from
-      ;; `haskell-indentation-find-indentations', but escapes the scope
-      ;; and aborts the opertaion before any moving happens
-      (let* ((cc (current-column))
-             (ci (haskell-indentation-current-indentation))
-             (inds (save-excursion
-                     (move-to-column ci)
-                     (or (haskell-indentation-find-indentations)
-                         '(0))))
-             (valid (memq ci inds))
-             (cursor-in-whitespace (< cc ci)))
+    ;; parse error is intentionally not cought here, it may come from
+    ;; `haskell-indentation-find-indentations', but escapes the scope
+    ;; and aborts the opertaion before any moving happens
+    (let* ((cc (current-column))
+           (ci (haskell-indentation-current-indentation))
+           (inds (save-excursion
+                   (move-to-column ci)
+                   (or (haskell-indentation-find-indentations)
+                       '(0))))
+           (valid (memq ci inds))
+           (cursor-in-whitespace (< cc ci)))
 
-        (if (and valid cursor-in-whitespace)
-            (move-to-column ci)
-          (haskell-indentation-reindent-to
-           (haskell-indentation-next-indentation ci inds 'nofail)
-           cursor-in-whitespace))
-        (setq haskell-indentation-dyn-last-direction 'right
-              haskell-indentation-dyn-first-position
-              (haskell-indentation-current-indentation)
-              haskell-indentation-dyn-last-indentations inds)))))
+      (if (and valid cursor-in-whitespace)
+          (move-to-column ci)
+        (haskell-indentation-reindent-to
+         (haskell-indentation-next-indentation ci inds 'nofail)
+         cursor-in-whitespace))
+      (setq haskell-indentation-dyn-last-direction 'right
+            haskell-indentation-dyn-last-indentations inds))))
 
 (defun haskell-indentation-indent-line-repeat ()
   "Cycle though indentation positions."
@@ -306,21 +293,22 @@ indentation points to the right, we switch going to the left."
               ci haskell-indentation-dyn-last-indentations 'nofail))
           ;; but failed, switch to left
           (setq haskell-indentation-dyn-last-direction 'left)
-          ;; and skip to the point where the user started pressing TABs.
-          ;; except if there are <= 2 indentation points, because this
-          ;; behavior is very confusing in that case
-          (when (< 2 (length haskell-indentation-dyn-last-indentations))
-            (haskell-indentation-reindent-to
-             haskell-indentation-dyn-first-position))
           (haskell-indentation-indent-line-repeat)))
       t))
    (t nil)))
 
-(defun haskell-indentation-indent-region (start end)
-  "Indent region from START to END."
-  (setq haskell-indentation-dyn-last-direction 'region)
-  (haskell-indentation-indent-rigidly start end 1)
-  (message "Press TAB or S-TAB again to indent the region more"))
+(defun haskell-indentation-indent-region (_start _end)
+  "This function does nothing.
+
+It is better to do nothing to indent region in Haskell than to
+break the semantics of indentation.  This function is used for
+`indent-region-function' because the default is to call
+`indent-line-function' on every line from START to END and that
+also produces catastrophic results.
+
+Someday we will have indent region that preserves semantics and
+fixes up only indentation."
+  nil)
 
 (defun haskell-indentation-indent-backwards ()
   "Indent the current line to the previous indentation point."
@@ -463,7 +451,20 @@ indentation points to the right, we switch going to the left."
   (let ((ppss (syntax-ppss)))
     (cond
      ((nth 3 ppss)
-      (haskell-indentation-first-indentation))
+      (if (save-excursion
+            (and (forward-line -1)
+                 (< (nth 8 ppss) (point))))
+          ;; if this string goes over more than one line we want to
+          ;; sync with the last line, not the first one
+          (list (save-excursion
+                  (forward-line -1)
+                  (current-indentation)))
+
+        (append
+         (haskell-indentation-first-indentation)
+         (list (save-excursion
+                 (goto-char (nth 8 ppss))
+                 (current-column))))))
      ((nth 4 ppss)
       (if (save-excursion
             (and (skip-syntax-forward "-")
@@ -519,6 +520,7 @@ indentation points to the right, we switch going to the left."
     ("type"    . haskell-indentation-data)
     ("newtype" . haskell-indentation-data)
     ("if"      . haskell-indentation-if)
+    ("case"    . haskell-indentation-case)
     ("let"     .
      ,(apply-partially 'haskell-indentation-phrase
                        '(haskell-indentation-declaration-layout
@@ -532,10 +534,6 @@ indentation points to the right, we switch going to the left."
     ("rec"     .
      ,(apply-partially 'haskell-indentation-with-starter
                        'haskell-indentation-expression-layout))
-    ("case"    .
-     ,(apply-partially 'haskell-indentation-phrase
-                       '(haskell-indentation-expression
-                         "of" haskell-indentation-case-layout)))
     ("\\"      .
      ,(apply-partially 'haskell-indentation-with-starter
                        'haskell-indentation-lambda-maybe-lambdacase))
@@ -576,7 +574,7 @@ indentation points to the right, we switch going to the left."
 
 (defun haskell-indentation-case-layout ()
   "Parse layout list with case expressions."
-  (haskell-indentation-layout #'haskell-indentation-case))
+  (haskell-indentation-layout #'haskell-indentation-case-item))
 
 (defun haskell-indentation-lambda-maybe-lambdacase ()
   "Parse lambda or lambda-case expression.
@@ -650,18 +648,22 @@ For example
 
 (defun haskell-indentation-data ()
   "Parse data or type declaration."
-  (haskell-indentation-with-starter
-   (lambda ()
-     (when (string= current-token "instance")
-       (haskell-indentation-read-next-token))
-     (haskell-indentation-type)
-     (cond ((string= current-token "=")
-            (haskell-indentation-with-starter
-             (apply-partially #'haskell-indentation-separated
-                              #'haskell-indentation-type "|" "deriving")))
-           ((string= current-token "where")
-            (haskell-indentation-with-starter
-             #'haskell-indentation-expression-layout nil))))))
+  (haskell-indentation-read-next-token)
+  (when (string= current-token "instance")
+    (haskell-indentation-read-next-token))
+  (haskell-indentation-type)
+  (cond ((eq current-token 'end-tokens)
+         (when (member following-token '("=" "where"))
+           (haskell-indentation-add-indentation current-indent)
+           (throw 'parse-end nil)))
+        ((string= current-token "=")
+         (haskell-indentation-with-starter
+          (lambda ()
+            (haskell-indentation-separated
+             #'haskell-indentation-expression "|" "deriving"))))
+        ((string= current-token "where")
+         (haskell-indentation-with-starter
+          #'haskell-indentation-expression-layout nil))))
 
 (defun haskell-indentation-import ()
   "Parse import declaration."
@@ -759,7 +761,7 @@ Skip the keyword or parenthesis." ; FIXME: better description needed
             ((equal current-token end)
              (haskell-indentation-read-next-token))))))
 
-(defun haskell-indentation-case-alternative ()
+(defun haskell-indentation-case-item-alternative ()
   "" ; FIXME
   (setq left-indent (current-column))
   (haskell-indentation-separated #'haskell-indentation-expression "," nil)
@@ -770,7 +772,7 @@ Skip the keyword or parenthesis." ; FIXME: better description needed
         ;; otherwise fallthrough
         ))
 
-(defun haskell-indentation-case ()
+(defun haskell-indentation-case-item ()
   "" ; FIXME
   (haskell-indentation-expression)
   (cond ((eq current-token 'end-tokens)
@@ -778,7 +780,7 @@ Skip the keyword or parenthesis." ; FIXME: better description needed
         ((string= current-token "|")
          (haskell-indentation-with-starter
           (apply-partially #'haskell-indentation-separated
-                           #'haskell-indentation-case-alternative "|" nil)
+                           #'haskell-indentation-case-item-alternative "|" nil)
           nil))
         ((string= current-token "->")
          (haskell-indentation-statement-right #'haskell-indentation-expression))
@@ -794,11 +796,10 @@ parser.  If parsing ends here, set indentation to left-indent."
     (haskell-indentation-add-left-indent)
     (haskell-indentation-add-indentation current-indent)
     (throw 'parse-end nil))
-  (let ((current-indent (current-column)))
-    (funcall parser)
-    (when (equal current-token "where")
-      (haskell-indentation-with-starter
-       #'haskell-indentation-expression-layout nil))))
+  (funcall parser)
+  (when (equal current-token "where")
+    (haskell-indentation-with-starter
+     #'haskell-indentation-expression-layout nil)))
 
 (defun haskell-indentation-guard ()
   "Parse \"guard\" statement."
@@ -809,15 +810,15 @@ parser.  If parsing ends here, set indentation to left-indent."
 (defun haskell-indentation-declaration ()
   "Parse function or type declaration."
   (haskell-indentation-separated #'haskell-indentation-expression "," nil)
-  (cond ((string= current-token "|")
-         (haskell-indentation-with-starter
-          (apply-partially #'haskell-indentation-separated
-                           #'haskell-indentation-guard "|" nil)
-          nil))
-        ((eq current-token 'end-tokens)
-         (when (member following-token '("|" "=" "::" ","))
-           (haskell-indentation-add-indentation current-indent)
-           (throw 'parse-end nil)))))
+  (when (string= current-token "|")
+    (haskell-indentation-with-starter
+     (apply-partially #'haskell-indentation-separated
+                      #'haskell-indentation-guard "|" nil)
+     nil))
+  (when (eq current-token 'end-tokens)
+   (when (member following-token '("|" "=" "::" ","))
+     (haskell-indentation-add-indentation current-indent)
+     (throw 'parse-end nil))))
 
 (defun haskell-indentation-layout (parser)
   "Parse layout list, where each layout item is parsed by parser."
@@ -833,35 +834,34 @@ parser.  If parsing ends here, set indentation to left-indent."
 
 (defun haskell-indentation-expression ()
   "Parse an expression until an unknown token is encountered."
-  (let ((current-indent (current-column)))
-    (catch 'return
-      (while t
-        (cond
-         ((memq current-token '(value operator))
-          (haskell-indentation-read-next-token))
-         ((eq current-token 'end-tokens)
-          (cond ((string= following-token "where")
-                 (haskell-indentation-add-where-pre-indent)) ; before a where
-                ((haskell-indentation-expression-token-p following-token)
-                 (haskell-indentation-add-indentation
-                  current-indent))) ; a normal expression
-          (throw 'return nil))
-         (t (let ((parser (assoc current-token
-                                 haskell-indentation-expression-list)))
-              (when (null parser)
-                (throw 'return nil)) ; not expression token, so exit
-              (funcall (cdr parser)) ; run parser
-              (when (and (eq current-token 'end-tokens)
-                         (string= (car parser) "let")
-                         (= haskell-indentation-layout-offset current-indent)
-                         (haskell-indentation-expression-token-p following-token))
-                ;; inside a layout, after a let construct
-                ;; for example: "do let a = 20"
-                (haskell-indentation-add-layout-indent)
-                (throw 'parse-end nil))
-              ;; after an 'open' expression such as 'if', exit
-              (unless (member (car parser) '("(" "[" "{" "case"))
-                (throw 'return nil)))))))))
+  (catch 'return
+    (while t
+      (cond
+       ((memq current-token '(value operator))
+        (haskell-indentation-read-next-token))
+       ((eq current-token 'end-tokens)
+        (cond ((string= following-token "where")
+               (haskell-indentation-add-where-pre-indent)) ; before a where
+              ((haskell-indentation-expression-token-p following-token)
+               (haskell-indentation-add-indentation
+                current-indent))) ; a normal expression
+        (throw 'return nil))
+       (t (let ((parser (assoc current-token
+                               haskell-indentation-expression-list)))
+            (when (null parser)
+              (throw 'return nil)) ; not expression token, so exit
+            (funcall (cdr parser)) ; run parser
+            (when (and (eq current-token 'end-tokens)
+                       (string= (car parser) "let")
+                       (= haskell-indentation-layout-offset current-indent)
+                       (haskell-indentation-expression-token-p following-token))
+              ;; inside a layout, after a let construct
+              ;; for example: "do let a = 20"
+              (haskell-indentation-add-layout-indent)
+              (throw 'parse-end nil))
+            ;; after an 'open' expression such as 'if', exit
+            (unless (member (car parser) '("(" "[" "{" "case"))
+              (throw 'return nil))))))))
 
 (defun haskell-indentation-separated (parser separator &optional stmt-separator)
   "Evaluate PARSER separated by SEPARATOR and STMT-SEPARATOR.
@@ -933,7 +933,9 @@ layout starts."
         (cond ((member current-token '(layout-item ";"))
                (haskell-indentation-read-next-token))
               ((eq current-token 'end-tokens)
-               (when (or (haskell-indentation-expression-token-p following-token)
+               (when (or (and
+                          (not (string= following-token "{"))
+                          (haskell-indentation-expression-token-p following-token))
                          (string= following-token ";")
                          (and (equal layout-indent 0)
                               (member following-token (mapcar #'car haskell-indentation-toplevel-list))))
@@ -955,7 +957,7 @@ layout starts."
          (haskell-indentation-with-starter
           (lambda ()
             (haskell-indentation-separated
-             #'haskell-indentation-case-alternative "|" nil))
+             #'haskell-indentation-case-item-alternative "|" nil))
           nil)
        (haskell-indentation-phrase-rest
         '(haskell-indentation-expression
@@ -963,66 +965,79 @@ layout starts."
           "else" haskell-indentation-expression))))
    nil))
 
+(defun haskell-indentation-case ()
+  "" ; FIXME
+  (haskell-indentation-with-starter
+   (lambda ()
+     (haskell-indentation-expression)
+     (when (string= current-token "of")
+       (haskell-indentation-with-starter
+        #'haskell-indentation-case-layout)))
+   nil))
+
 (defun haskell-indentation-phrase (phrase)
   "" ; FIXME
   (haskell-indentation-with-starter
    (apply-partially #'haskell-indentation-phrase-rest phrase)
    nil))
 
-(defun haskell-indentation-phrase-rest (phrase)
+(defun haskell-indentation-phrase-rest (phrase1)
   "" ; FIXME
-  (let ((starter-line parse-line-number))
-    (let ((current-indent (current-column)))
-      (funcall (car phrase)))
-    (cond
-     ((eq current-token 'end-tokens)
-      (cond ((null (cdr phrase))) ;; fallthrough
-            ((equal following-token (cadr phrase))
-             (haskell-indentation-add-indentation starter-indent)
-             (throw 'parse-end nil))
-            ((string= (cadr phrase) "in")
-             (when (= left-indent layout-indent)
-               (haskell-indentation-add-layout-indent)
-               (throw 'parse-end nil)))
-            (t (throw 'parse-end nil))))
-     ((null (cdr phrase)))
-     ((equal (cadr phrase) current-token)
-      (let* ((on-new-line (= (current-column)
-                             (haskell-indentation-current-indentation)))
-             (lines-between (- parse-line-number starter-line))
-             (left-indent (if (<= lines-between 0)
-                              left-indent
-                            starter-indent)))
-        (haskell-indentation-read-next-token)
-        (when (eq current-token 'end-tokens)
-          (cond ((member (cadr phrase) '("then" "else"))
-                 (haskell-indentation-add-indentation
-                  (+ starter-indent haskell-indentation-ifte-offset)))
+  (while phrase1
+    (let ((starter-line parse-line-number)
+          (phrase phrase1))
+      (setq phrase1 nil)
+      (let ((current-indent (current-column)))
+        (funcall (car phrase)))
+      (cond
+       ((eq current-token 'end-tokens)
+        (cond ((null (cdr phrase))) ;; fallthrough
+              ((equal following-token (cadr phrase))
+               (haskell-indentation-add-indentation starter-indent)
+               (throw 'parse-end nil))
+              ((string= (cadr phrase) "in")
+               (when (= left-indent layout-indent)
+                 (haskell-indentation-add-layout-indent)
+                 (throw 'parse-end nil)))
+              (t (throw 'parse-end nil))))
+       ((null (cdr phrase)))
+       ((equal (cadr phrase) current-token)
+        (let* ((on-new-line (= (current-column)
+                               (haskell-indentation-current-indentation)))
+               (lines-between (- parse-line-number starter-line))
+               (left-indent (if (<= lines-between 0)
+                                left-indent
+                              starter-indent)))
+          (haskell-indentation-read-next-token)
+          (when (eq current-token 'end-tokens)
+            (cond ((member (cadr phrase) '("then" "else"))
+                   (haskell-indentation-add-indentation
+                    (+ starter-indent haskell-indentation-ifte-offset)))
 
-                ((member (cadr phrase) '("in" "->"))
-                 ;; expression ending in another expression
-                 (when (or (not haskell-indentation-indent-leftmost)
-                           (eq haskell-indentation-indent-leftmost 'both))
-                   (haskell-indentation-add-indentation
-                    (+ starter-indent haskell-indentation-starter-offset)))
-                 (when haskell-indentation-indent-leftmost
-                   (haskell-indentation-add-indentation
-                    (if on-new-line
-                        (+ left-indent haskell-indentation-starter-offset)
-                      left-indent))))
-                (t
-                 (when (or (not haskell-indentation-indent-leftmost)
-                           (eq haskell-indentation-indent-leftmost 'both))
-                   (haskell-indentation-add-indentation
-                    (+ starter-indent haskell-indentation-starter-offset)))
-                 (when haskell-indentation-indent-leftmost
-                   (haskell-indentation-add-indentation
-                    (if on-new-line
-                        (+ left-indent haskell-indentation-starter-offset)
-                      left-indent)))))
-          (throw 'parse-end nil))
-        (haskell-indentation-phrase-rest (cddr phrase))))
-     ((string= (cadr phrase) "in")))))
+                  ((member (cadr phrase) '("in" "->"))
+                   ;; expression ending in another expression
+                   (when (or (not haskell-indentation-indent-leftmost)
+                             (eq haskell-indentation-indent-leftmost 'both))
+                     (haskell-indentation-add-indentation
+                      (+ starter-indent haskell-indentation-starter-offset)))
+                   (when haskell-indentation-indent-leftmost
+                     (haskell-indentation-add-indentation
+                      (if on-new-line
+                          (+ left-indent haskell-indentation-starter-offset)
+                        left-indent))))
+                  (t
+                   (when (or (not haskell-indentation-indent-leftmost)
+                             (eq haskell-indentation-indent-leftmost 'both))
+                     (haskell-indentation-add-indentation
+                      (+ starter-indent haskell-indentation-starter-offset)))
+                   (when haskell-indentation-indent-leftmost
+                     (haskell-indentation-add-indentation
+                      (if on-new-line
+                          (+ left-indent haskell-indentation-starter-offset)
+                        left-indent)))))
+            (throw 'parse-end nil))
+          (setq phrase1 (cddr phrase))))
+       ((string= (cadr phrase) "in"))))))
 
 (defun haskell-indentation-add-indentation (indent)
   "" ; FIXME
